@@ -32,7 +32,6 @@ G_BEGIN_DECLS
 #define PANGO_IS_FONT_MAP(object)        (G_TYPE_CHECK_INSTANCE_TYPE ((object), PANGO_TYPE_FONT_MAP))
 
 typedef struct _PangoContext PangoContext;
-typedef struct _PangoFontMap PangoFontMap;
 
 GType         pango_font_map_get_type       (void) G_GNUC_CONST;
 PangoFont *   pango_font_map_load_font     (PangoFontMap                 *fontmap,
@@ -46,7 +45,6 @@ void          pango_font_map_list_families (PangoFontMap                 *fontma
 					    PangoFontFamily            ***families,
 					    int                          *n_families);
 
-
 #ifdef PANGO_ENABLE_BACKEND
 
 #define PANGO_FONT_MAP_CLASS(klass)      (G_TYPE_CHECK_CLASS_CAST ((klass), PANGO_TYPE_FONT_MAP, PangoFontMapClass))
@@ -63,7 +61,9 @@ struct _PangoFontMap
 struct _PangoFontMapClass
 {
   GObjectClass parent_class;
-  
+
+  /*< public >*/
+
   PangoFont *   (*load_font)     (PangoFontMap               *fontmap,
 				  PangoContext               *context,
 				  const PangoFontDescription *desc);
@@ -74,8 +74,20 @@ struct _PangoFontMapClass
 				  PangoContext               *context,
 				  const PangoFontDescription *desc,
 				  PangoLanguage              *language);
+
+  const char *shape_engine_type;
+
+  /*< private >*/
+
+  /* Padding for future expansion */
+  void (*_pango_reserved1) (void);
+  void (*_pango_reserved2) (void);
+  void (*_pango_reserved3) (void);
+  void (*_pango_reserved4) (void);
 };
 
+const char   *pango_font_map_get_shape_engine_type (PangoFontMap *fontmap);
+
 #endif /* PANGO_ENABLE_BACKEND */
 
 G_END_DECLS
