@@ -34,8 +34,10 @@
 #include <pango/pango-fontset.h>
 #include <pango/pango-glyph.h>
 #include <pango/pango-glyph-item.h>
+#include <pango/pango-gravity.h>
 #include <pango/pango-item.h>
 #include <pango/pango-layout.h>
+#include <pango/pango-matrix.h>
 #include <pango/pango-renderer.h>
 #include <pango/pango-script.h>
 #include <pango/pango-tabs.h>
